/* mC generator koda na HAJ */

%{
  #include <stdio.h>
  #include "defs.h"
  #include "symtab.h"
  #include "stack.h"
  #include "sem.h"
  #include "codegen.h"

  extern int yylineno;
  char char_buffer[CHAR_BUFFER_LENGTH];
  
  int error_count = 0;
  int var_num = 0;
  int function_index = -1;
  int function_call_index = -1;
  int lab_num;
  int false_lab_num = -1;
  stack label_stack;
  FILE *output;
%}

%union {
  int i;
  char *s;
}


%token _TYPE
%token _IF
%token _ELSE
%token _RETURN
%token <s> _ID
%token <s> _INT_NUMBER
%token _LPAREN
%token _RPAREN
%token _LBRACKET
%token _RBRACKET
%token _ASSIGN
%token _SEMICOLON
%token <i> _AROP
%token <i> _RELOP

%type <i> type variable num_exp exp constant 
%type <i> function_call rel_exp

%%

program
  : function_list
      { check_main(); }
  ;

function_list
  : function
  | function_list function
  ;

function
  : type _ID
      {
        function_index = try_to_insert_id($2,FUNCTION,$1);
        var_num = 0;

        code("\n%s:", $2);
        code("\n\t\t\tPUSH\t%%14");
        code("\n\t\t\tMOV \t%%15,%%14");
      }
    _LPAREN _RPAREN body
      {
        if (get_last_element() > function_index)
          clear_symbols(function_index + 1);

        gen_str_lab($2,"_exit");
        code("\n\t\t\tMOV \t%%14,%%15");
        code("\n\t\t\tPOP \t%%14");
        code("\n\t\t\tRET");
      }
  ;

type
  : _TYPE
      {  $$ = INT_TYPE; }
  ;

body
  : _LBRACKET variable_list
      {
        code("\n\t\t\tSUBU\t%%15,$%d,%%15", 4*var_num);
        gen_str_lab(get_name(function_index), "_body");
      }
    statement_list _RBRACKET

  | _LBRACKET
      { gen_str_lab(get_name(function_index), "_body"); }
    statement_list _RBRACKET
  ;

variable_list
  : variable _SEMICOLON
  | variable_list variable _SEMICOLON
  ;

variable
  : type _ID
      {
        $$ = try_to_insert_id($2, LOCAL_VAR, $1);
        var_num++;
        set_attribute($$, var_num);
      }
  ;

statement_list
  : /* empty */
  | statement_list statement
  ;

statement
  : compound_statement
  | assignment_statement
  | if_statement
  | return_statement
  ;

compound_statement
  : _LBRACKET statement_list _RBRACKET
  ;

assignment_statement
  : _ID _ASSIGN num_exp _SEMICOLON
      {
        int i = -1;
        if( (i = lookup_id($1, LOCAL_VAR)) == -1 )
          printerror("invalid lvalue in assignment");
        if(!check_types(i, $3))
          printerror("incompatible types in assignment");
        gen_mov($3, i);
      }
  ;

num_exp
  : exp

  | num_exp _AROP exp
      {
        if(!check_types($1, $3))
          printerror("invalid operands: arithmetic operation");
        $$ = gen_arith($2, $1, $3);
      }
  ;

exp
  : constant

  | _ID
      {
        if( ($$ = lookup_id($1, LOCAL_VAR)) == -1)
          printerror("'%s' undeclared", $1);
      }

  | function_call
      {
        $$ = take_reg();
        gen_mov(FUNCTION_REG, $$);
      }
  
  | _LPAREN num_exp _RPAREN
      { $$ = $2; }
  ;

constant
  : _INT_NUMBER
      { $$ = try_to_insert_constant($1, INT_TYPE); }
  ;

function_call
  : _ID _LPAREN _RPAREN
      {
        if((function_call_index = lookup_id($1, FUNCTION))==-1)
          printerror("'%s' is not a function", $1);
        
        gen_function_call(function_call_index);
        set_register_type(FUNCTION_REG, 
          get_type(function_call_index)); 
        $$ = FUNCTION_REG;
      }
  ;

if_statement
  : if_part
      { gen_num_lab("exit", lab_num); }

  | if_part _ELSE
      { push(&label_stack, lab_num); }
    statement
      { gen_num_lab("exit", pop(&label_stack)); }
  ;

if_part
  : _IF _LPAREN
      {
        lab_num = ++false_lab_num;
        gen_num_lab("if", lab_num);
      }
    rel_exp
      {
        const char* opposite_jumps[2]={"JNE ", "JEQ "};
        code("\n\t\t\t%s\t@false%d", 
          opposite_jumps[$4], false_lab_num);
        gen_num_lab("true", lab_num);
        push(&label_stack, false_lab_num);
        push(&label_stack, lab_num);
      }
    _RPAREN statement
      {
        lab_num = pop(&label_stack);
        code("\n\t\t\tJMP\t@exit%d", lab_num);
        gen_num_lab("false", pop(&label_stack));
      }
  ;

rel_exp
  : num_exp _RELOP num_exp
      {
        if(!check_types($1, $3))
          printerror("invalid operands: relational operator");
        gen_cmp($1, $3);
        $$ = $2;
      }
  ;

return_statement
  : _RETURN num_exp _SEMICOLON
      {
        if(!check_types(function_index, $2))
          printerror("incompatible types in return");

        // vrednost return izraza se prebaci u registar %13
        gen_mov($2, FUNCTION_REG);
        code("\n\t\t\tJMP\t@%s_exit",get_name(function_index));        
      }
  ;


%%


int yyerror(char *s) {
  fprintf(stderr, "\nline %d: ERROR: %s", yylineno, s);
  error_count++;
  return 0;
}

void warning(char *s) {
  fprintf(stderr, "\nline %d: WARNING: %s", yylineno, s);
}

int main() {
  printf("\nSTART\n");
  init_symtab();
  init_stack(&label_stack);
  output = fopen("output.asm", "w+");

  yyparse();

  clear_symtab();
  fclose(output);
  printf("\nSTOP\n");
  
  if(error_count)
    remove("output.asm");
  return error_count;
}

